ThePow:=function(a, eExp)
  local eVal, i;
  eVal:=1;
  for i in [1..eExp]
  do
    eVal:=eVal*a;
  od;
  return eVal;
end;

# compute P_i(x) 
KrawtchukPolynomial:=function(n,i,x)
  local TheSum, j, eVal1, eVal2, eVal3;
  TheSum:=0;
  for j in [0..i]
  do
    eVal1:=ThePow(-1, j);
    eVal2:=NrCombinations([1..x],j);
    eVal3:=NrCombinations([1..n-x],i-j);
    TheSum:=TheSum + eVal1*eVal2*eVal3;
  od;
  return TheSum;
end;



# The Krawtchuk polynomials are determined by the expression
# sum_{i=0}^n P_i(x) z^i = (1+z)^{n-x} (1-z)^x
# This gives the following expressions
# (1+z)^{n-x} (1-z)^x = [ sum_j (n-x,j) z^j ] [ sum_k (-1)^k (x,k) z^k ]
#                     = sum_i P_i(x) z^i
# Therefore we have
# P_i(x) = sum_{j=0}^i (-1)^j (x,j) (n-x,i-j) 
# and also
# ---P_i(0) = (n,i)
# ---sum P_i(x) = | 2^n if x=0
#                 | 0   if x>0
# ---P_0(x) = 1
#
# Relations that we have is
# A_0 = 1 and B_0 = 1.
#
# And also B_i = 2^(-k) sum_{j=0}^n A_j Pi(j)
#
# The inequality B_i >= 0 imply that
# 
TestDelsarteFeasibilitySystem:=function(n, k, d)
  local ListIneq, i, eEqua, eIneq, j, TheSpace, FACred, eFAC, ListSymb, DoType3;
  ListIneq:=[];
  ListSymb:=[];
  # equations A_i = 0 for i=1, ..., d-1
  for i in [1..d-1]
  do
    eEqua:=ListWithIdenticalEntries(n+1,0);
    eEqua[i+1]:=1;
    Add(ListIneq, ShallowCopy(eEqua));
    Add(ListSymb, Concatenation("A", String(i), " >= 0"));
    #
    Add(ListIneq, -ShallowCopy(eEqua));
    Add(ListSymb, Concatenation("A", String(i), " <= 0"));
  od;
  # inequalities A_i >= 0
  for i in [d..n]
  do
    eIneq:=ListWithIdenticalEntries(n+1,0);
    eIneq[i+1]:=1;
    Add(ListIneq, eIneq);
    Add(ListSymb, Concatenation("A", String(i), " >= 0"));
  od;
  #
  # inequality B_i >= 0 for i >= 1.
  # rewritten as sum Pi(j) Aj >= - (n i), i.e. type 1
  #
  for i in [1..n]
  do
    eIneq:=ListWithIdenticalEntries(n+1,0);
    for j in [0..n]
    do
      eIneq[1+j]:=KrawtchukPolynomial(n, i,j);
    od;
    Add(ListIneq, eIneq);
    Add(ListSymb, Concatenation("B", String(i), " >= 0"));
  od;
  # inequalities A_i + B_i <= (n,i) for i>=1, i.e. type 2
  for i in [1..n]
  do
    eIneq:=ListWithIdenticalEntries(n+1,0);
    eIneq[1]:=NrCombinations([1..n],i);
    eIneq[1+i]:=-1;
    for j in [0..n]
    do
      eIneq[1+j]:=eIneq[1+j] - 2^(-k)*KrawtchukPolynomial(n, i, j);
    od;
    Add(ListIneq, eIneq);
    Add(ListSymb, Concatenation("A", String(i), " + B", String(i), " <= (n ", String(i), ")"));
  od;
  # inequalities of type 3
  DoType3:=true;
#  DoType3:=false;
  if DoType3=true then
    for i in [1..n]
    do
      eIneq:=ListWithIdenticalEntries(n+1,0);
      eIneq[1+i]:=-2^k;
      for j in [1..n]
      do
        eIneq[1+j]:=eIneq[1+j] - (KrawtchukPolynomial(n, i,j) - KrawtchukPolynomial(n, i,0));
      od;
      Add(ListIneq, eIneq);
      Add(ListSymb, "inequality of type 3");
    od;
  fi;
  #
  # Dimension equation:
  # equation 2^k -1 = sum_{i>=1} A_i
  # This is also obtained from 
  # equality B_0 = 1
  # rewritten as
  # 2^k - 1 = sum_{j=1}^n Aj P0(j)
  #         = sum_{j=1}^n Aj
  #
  eIneq:=ListWithIdenticalEntries(n+1,0);
  eIneq[1]:=2^k-1;
  for i in [1..n]
  do
    eIneq[1+i]:=-1;
  od;
  Add(ListIneq, ShallowCopy(eIneq));
  Add(ListSymb, "inequality 2^k >= sum_i A_i");
  Add(ListIneq, -ShallowCopy(eIneq));
  Add(ListSymb, "inequality 2^k <= sum_i A_i");
  #
  # affine convex body
  #
  eIneq:=ListWithIdenticalEntries(n+1,0);
  eIneq[1]:=1;
  Add(ListIneq, eIneq);
  Add(ListSymb, "affine convex body");
  #
  # Now the linear programming part
  # 
  TheSpace:=LinearDeterminedByInequalities(ListIneq);
  if Length(TheSpace)=0 then
    return rec(answer:="TheSpace is empty", ListIneq:=ListIneq, ListSymb:=ListSymb);
  fi;
  FACred:=[];
  for eIneq in ListIneq
  do
    eFAC:=List(TheSpace, x->x*eIneq);
    Add(FACred, eFAC);
  od;
  return rec(answer:="may be feasible",
             TheSpace:=TheSpace,
             FACred:=FACred,
             ListIneq:=ListIneq,
             ListSymb:=ListSymb);
end;

GetValueSimpleLP:=function(n,k,d)
  local FileSave, RecTest, TheDim, TheResult, RecSave;
  FileSave:=Concatenation("ResultLCD_old/LCDtest_", String(n), "_", String(k), "_", String(d));
  if IsExistingFile(FileSave) then
    return ReadAsFunction(FileSave)();
  fi;
  RecTest:=TestDelsarteFeasibilitySystem(n, k, d);
  if RecTest.answer="TheSpace is empty" then
    TheDim:=0;
  else
    TheDim:=Length(RecTest.TheSpace);
  fi;
  if TheDim>0 then
    TheResult:=true;
  else
    TheResult:=false;
  fi;
  RecSave:=rec(n:=n, k:=k, d:=d, TheDim:=TheDim, result:=TheResult);
  SaveDataToFile(FileSave, RecSave);
  return RecSave;
end;


GetMaximumValueD_SimpleLP:=function(n,k)
  local d, test1, test2, testFinal;
  if GetValueSimpleLP(n,k,1).result=false then
    return 0;
  fi;
  for d in [1..n-1]
  do
    test1:=GetValueSimpleLP(n,k,d).result;
    test2:=GetValueSimpleLP(n,k,d+1).result;
    if test1=true and test2=false then
      return d;
    fi;
  od;
  testFinal:=GetValueSimpleLP(n,k,n).result;
  if testFinal<>true then
    Error("Consistency error\n");
  fi;
  return n;
end;



#
# Now the next part of the code
#



DetermineLastValue:=function(n, k)
  local GetFile, d, eFile1, eFile2, eRec1, eRec2;
  GetFile:=function(d)
    return Concatenation("ResultLCD_cdd/out", String(n), "_", String(k), "_", String(d), "_cdd");
  end;
  for d in [2..n-1]
  do
    eFile1:=GetFile(d);
    eFile2:=GetFile(d+1);
    if IsExistingFile(eFile1)=true and IsExistingFile(eFile2)=true then
      eRec1:=ReadAsFunction(eFile1)();
      eRec2:=ReadAsFunction(eFile2)();
      if eRec1.TheDim>0 and eRec2.TheDim=0 then
        return rec(success:=true, MaxD:=d);
      fi;
    fi;
  od;
  return rec(success:=false);
end;



DirectComputationSpecificValue:=function(n,k,d)
  local eComm, eFileOut, option, strMeth, eCommand, eRec;
  eComm:="JointLCDcodeBound";
  eFileOut:=Concatenation("ResultLCD_cdd/out", String(n), "_", String(k), "_", String(d), "_cdd");
  if IsExistingFile(eFileOut)=false then
    option:=0;
    strMeth:="cdd";
    eCommand:=Concatenation(eComm, " ", String(n), " ", String(k), " ", String(d), " ", strMeth, " ", eFileOut, " ", String(option));
    Print("eCommand=", eCommand, "\n");
    Exec(eCommand);
  fi;
  #
  if IsExistingFile(eFileOut)=false then
    Error("Missing file");
  fi;
  eRec:=ReadAsFunction(eFileOut)();
  return eRec;
end;




ComputeSpecificValue:=function(n,k,d)
  local eComm, eFileOut, eCommand, res, option, eRec, strMeth;
  if k>n or d>n then
    Print("n=", n, " k=", k, " d=", d, "\n");
    Error("Logical error in the input");
  fi;
  if d=1 then
    return rec(n:=n, k:=k, d:=d, TheDim:=1, result:=true);
  fi;
  if d=2 then
    res:=n mod 2;
    if res=0 then
      if k<=n-2 then
        return rec(n:=n, k:=k, d:=d, TheDim:=1, result:=true);
      else
        return rec(n:=n, k:=k, d:=d, TheDim:=0, result:=false);
      fi;
    fi;
    if res=1 then
      if k<=n-1 then
        return rec(n:=n, k:=k, d:=d, TheDim:=1, result:=true);
      else
        return rec(n:=n, k:=k, d:=d, TheDim:=0, result:=false);
      fi;
    fi;
  fi;
  if k=1 then
    res:=d mod 2;
    if res = 1 or d+1 <= n then
      return rec(n:=n, k:=k, d:=d, TheDim:=1, result:=true);
    else
      return rec(n:=n, k:=k, d:=d, TheDim:=0, result:=false);
    fi;
  fi;
  #
  return DirectComputationSpecificValue(n,k,d);
end;


ComputeSpecificValue_of_MaxD:=function(n,k)
  local d, eRec1, eRec2, ListRecD, ListDim;
  if n=1 then
    return rec(success:=true, MaxD:=1);
  fi;
  for d in [1..n-1]
  do
    eRec1:=ComputeSpecificValue(n,k,d);
    eRec2:=ComputeSpecificValue(n,k,d+1);
    if eRec1.TheDim>0 and eRec2.TheDim=0 then
      return rec(success:=true, MaxD:=d);
    fi;
  od;
  ListRecD:=List([1..n], d->ComputeSpecificValue(n,k,d));
  ListDim:=List(ListRecD, x->x.TheDim);
  if Minimum(ListDim)>0 then
    return rec(success:=true, MaxD:=n);
  fi;
  if Maximum(ListDim)=0 then
    return rec(success:=true, MaxD:=0);
  fi;
  Print("n=", n, " k=", k, "\n");
  Error("Failed to find matching entry");
end;


GeneralGetValueMaxD:=function(n,k)
  local eRec, eRecComp;
  eRec:=DetermineLastValue(n, k);
  if eRec.success=true then
    return eRec.MaxD;
  fi;
  eRecComp:=ComputeSpecificValue_of_MaxD(n,k);
  return eRecComp.MaxD;
end;

GeneralIsFeasibleTriple:=function(n,k,d)
  local MaxD;
  MaxD:=GeneralGetValueMaxD(n,k);
  if d<=MaxD then
    return true;
  fi;
  return false;
end;



GetKSetFeasible:=function(n,d)
  local TheSetFeasible, k;
  TheSetFeasible:=[];
  for k in [1..n]
  do
    if GeneralIsFeasibleTriple(n,k,d) then
      Add(TheSetFeasible, k);
    fi;
  od;
  return TheSetFeasible;
end;




PrintArrayUniqueWeightEnumerator:=function(FileSave, eRec)
  local output, len, PreVect, eVect, i;
  output:=OutputTextFile(FileSave, true);
  AppendTo(output, "Print information for case\n");
  AppendTo(output, "n=", eRec.n, " k=", eRec.k, " d=", eRec.d, "\n");
  len:=Length(eRec.TypeVar);
  PreVect:=eRec.LinSpaceFull[1];
  eVect:=PreVect/PreVect[1];
  AppendTo(output, "non-zero indices of w(i,j,k,l):\n");
  for i in [1..len]
  do
    if eRec.TypeVar[i]=1 and eVect[i]<>0 then
      AppendTo(output, "w(", eRec.ListVarInfo[i], ") = ", eVect[i], "\n");
    fi;
  od;
  AppendTo(output, "Field A:\n");
  for i in [1..len]
  do
    if eRec.TypeVar[i]=2 then
      AppendTo(output, "A", eRec.ListVarInfo[i], " = ", eVect[i], "\n");
    fi;
  od;
  AppendTo(output, "Field B:\n");
  for i in [1..len]
  do
    if eRec.TypeVar[i]=3 then
      AppendTo(output, "B", eRec.ListVarInfo[i], " = ", eVect[i], "\n");
    fi;
  od;
  CloseStream(output);
end;


GetIntegralPoints:=function(eRec)
  local NSP, TheSpace, nbVect, dim, FirstCol, eRecGcd, eBasisVect, iVect, ListDiff, eVect, eNewVect, SpanBasis, TheBasis, FAC, eIneq, fIneq, RecInt, ListPoint, eVectI, FACred, EXT, NSPred, TheDim, UpperBoundPoss, eLow, eUpp, ListVal, iDim, ListLow, ListUpp, ListEXTnorm, ListPointIter, ListPointZsolve, DoZsolve;
#  Print("nbVect=", Length(eRec.LinSpaceFull), "\n");
#  Print("dim=", Length(eRec.LinSpaceFull[1]), "\n");
  NSP:=NullspaceMat(TransposedMat(eRec.LinSpaceFull));
#  Print("We have NSP\n");
  NSPred:=List(NSP, RemoveFraction);
  TheSpace:=NullspaceIntMat(TransposedMat(RemoveFractionMatrix(NSP)));
#  Print("We have TheSpace\n");
  nbVect:=Length(TheSpace);
  #
  dim:=Length(TheSpace[1]);
#  Print("dim=", dim, "\n");
  FirstCol:=List(TheSpace, x->x[1]);
  eRecGcd:=GcdVector(FirstCol);
#  Print("We have eRecGcd\n");
  if eRecGcd.TheGcd<>1 then
    return rec(reply:="no integral points");
  fi;
  eBasisVect:=ListWithIdenticalEntries(dim,0);
  for iVect in [1..nbVect]
  do
    eBasisVect:=eBasisVect + eRecGcd.ListCoef[iVect]*TheSpace[iVect];
  od;
#  Print("We have eBasisVect\n");
  ListDiff:=[];
  for eVect in TheSpace
  do
    eNewVect:=eVect - eVect[1]*eBasisVect;
    Add(ListDiff, eNewVect);
  od;
  SpanBasis:=GetZbasis(ListDiff);
  TheDim:=Length(SpanBasis);
#  Print("We have SpanBasis\n");
  TheBasis:=Concatenation([eBasisVect], SpanBasis);
  #
  FAC:=[];
  for eIneq in eRec.ListIneq
  do
    fIneq:=List(TheBasis, x->x*eIneq);
    AddSet(FAC, RemoveFraction(fIneq));
  od;
  Print("We have FAC |FAC|=", Length(FAC), "\n");
#  Print(NullMat(5));
  FACred:=RemoveRedundancyByDualDescription(FAC);
#  Print("We have FACred |FACred|=", Length(FACred), "\n");
  EXT:=DualDescription(FACred);
#  Print("We have EXT |EXT|=", Length(EXT), "\n");
  ListEXTnorm:=List(EXT, x->x/x[1]);
  ListLow:=[];
  ListUpp:=[];
  UpperBoundPoss:=1;
  for iDim in [1..TheDim]
  do
    ListVal:=List(ListEXTnorm, x->x[iDim+1]);
    eLow:=UpperInteger(Minimum(ListVal));
    eUpp:=LowerInteger(Maximum(ListVal));
    Add(ListLow, eLow);
    Add(ListUpp, eUpp);
    UpperBoundPoss:=UpperBoundPoss*(eUpp+1-eLow);
  od;
#  Print("UpperBoundPoss=", UpperBoundPoss, "\n");
  ListPointIter:=GetIntegralPointsIteration(FACred, EXT);
  DoZsolve:=false;
  if DoZsolve then
    RecInt:=RunZsolve(FACred);
    ListPointZsolve:=List(RecInt.TheINH, x->Concatenation([1],x));
    if Set(ListPointZsolve)<>Set(ListPointIter) then
      Print("You have some debugging work to do\n");
      Print("|RecInt.TheINH|=", Length(RecInt.TheINH), "\n");
      Print("|ListPointIter|=", Length(ListPointIter), "\n");
      Error("Please correct");
    fi;
  fi;
  ListPoint:=[];
  for eVect in ListPointIter
  do
    eVectI:=eVect*TheBasis;
    Add(ListPoint, eVectI);
  od;
  return rec(FACred:=FACred, EXT:=EXT, ListPoint:=ListPoint);
end;


